import click
import json
import sys
import traceback

from lhub_extractors import util, params
from lhub_extractors import feature_extractor
from dataclasses_json import dataclass_json
from dataclasses import dataclass
from typing import List

@dataclass
@dataclass_json
class FeatureExtractorExecutionResult:
    entrypoint: str
    result: str
    errors: List[str]

@dataclass
@dataclass_json
class FeatureExtractorExecutionResults:
    results: List[FeatureExtractorExecutionResult]
    errors: List[str]

node_metadata = {}
required_columns = {}
node_output_table = []

def read_input_data():
    global node_metadata
    global node_output_table

    as_dict_1 = json.loads(sys.stdin.readline())

    if 'node_metadata' in as_dict_1:
        node_metadata = as_dict_1['node_metadata']

    for data in sys.stdin.readlines():
        as_dict = json.loads(data)

        if 'row' in as_dict:
            row = as_dict['row']
            node_output_table.append(row)

def run_extractor(entrypoint, entrypoint_fn):
    global node_metadata
    global node_output_table

    type_mapper = params.get_input_converter(entrypoint_fn)
    column_to_type_map = params.get_input_type_map(entrypoint_fn)

    errors = []
    required_columns = list(column_to_type_map.keys())
    node_output_table_columns = node_metadata['columns']
    feature_extractor_node_output_table = []

    if required_columns <= node_output_table_columns:
        for row in node_output_table:
            try:
                def coerce(column, column_data):
                    try:
                        return type_mapper[column](row, column)
                    except ValueError:
                        errors.append(
                            f"Invalid value: {column_data} for column {column}. Expected data of type: {column_to_type_map[column]}"
                        )
                        raise EndOfLoop()

                mapped_row = {}
                for column, column_data in row.items():
                    if column in required_columns:
                        mapped_column_data = coerce(column, column_data)
                        mapped_row[column] = mapped_column_data
                feature_extractor_node_output_table.append(mapped_row)

            except EndOfLoop:
                break

            except Exception:
                error = json.dumps({
                 "error": traceback.format_exc(),
                 "data": row
                })
                errors.append(error)
                break
    else:
        errors.append(f"Required columns: {required_columns} did not exist in the node output table columns: {node_output_table_columns}")

    result = {}

    if not errors:
        try:
            result = entrypoint_fn(node_metadata, feature_extractor_node_output_table)
            if result:
                return FeatureExtractorExecutionResult(
                    entrypoint = entrypoint,
                    result = result,
                    errors = errors
                )
        except Exception:
            error = json.dumps({
                "error": traceback.format_exc(),
                "data": row
            })
            errors.append(error)

    return FeatureExtractorExecutionResult(
        entrypoint = entrypoint,
        result = result,
        errors = errors
    )

# expects a list of feature_extractors to execute
def run_extractors(entrypoints):
    results = []
    errors = []
    for entrypoint in entrypoints.split(","):
        try:
            entrypoint_fn = feature_extractor.all().get(entrypoint).function
            result = run_extractor(entrypoint, entrypoint_fn)
            results.append(result)
        except Exception as ex:
            errors.append(
                json.dumps({
                    "entrypoint": entrypoint,
                    "error": ex
                })
            )

    feature_extractor_results = FeatureExtractorExecutionResults(
        results = results,
        errors = errors
    )

    util.print_result(feature_extractor_results.to_json())

class EndOfLoop(Exception):
    pass

@click.command()
@click.option("--entrypoints", "-e", required=True)
def main(entrypoints):
    errors = util.import_workdir()
    if errors:
        util.hard_exit_from_instantiation(str(errors[0]))
    read_input_data()
    run_extractors(entrypoints)

if __name__ == "__main__":
    main()


from lhub_extractors import util

class feature_extractor:
    extractors = {}

    def add_extractor(self, f):
        self.function = f
        self.entrypoint = f"{f.__module__}.{f.__name__}"
        if not self.name:
            self.name = f.__name__
        self.extractors[self.entrypoint] = self

    def __init__(self, name=None):
        self.name = None
        self.function = None
        self.entrypoint = None
        # if the first argument is callable, the first argument is actually the function
        if name and callable(name):
            self.add_extractor(name)
        else:
            self.name = name

    def __call__(self, f, *args, **kwargs):
        if self.function:
            return self.function(f, *args, **kwargs)
        else:
            if not f:
                util.invalid_extractor(
                    code="invalid_extractor",
                    error=f"@feature_extractor: {self.name}  must be called with a function"
                )
            else:
                self.add_extractor(f)
                return f

    @classmethod
    def all(cls):
        return cls.extractors


from dataclasses_json import dataclass_json
from dataclasses import dataclass
from docstring_parser import parse
from typing import List

import inspect
import json

from lhub_extractors.util import print_result, import_workdir, InvalidExtractor
from lhub_extractors import feature_extractor

@dataclass
@dataclass_json
class FeatureExtractorColumn:
    name: str
    data_type: str
    description: str
    errors: List[str]

@dataclass
@dataclass_json
class FeatureExtractorMetadata:
    feature_extractor_name: str
    feature_extractor_description: str
    feature_extractor_columns: List[FeatureExtractorColumn]
    feature_extractor_entrypoint: str
    errors: List[str]

def generate_metadata():
    toplevel_errors = []
    import_errors = import_workdir()
    for error in import_errors:
        if isinstance(error, SyntaxError):
            toplevel_errors.append(f"Failed to import module: {error}")
        elif isinstance(error, ImportError):
            toplevel_errors.append(
                f"Failed to import module (did you run bundle-extractors?): {error}"
            )
        elif isinstance(error, InvalidExtractor):
            toplevel_errors.append(
                f"Extractor was invalid: {error}"
            )
        else:
            toplevel_errors.append(
                f"Failed to import module (Unexpected error): {error}"
            )

    extractors = []
    for extractor in feature_extractor.all().values():
        extractor_function = extractor.function

        # Parse the docstring
        docs = extractor_function.__doc__
        parsed = parse(docs)

        errors = []
        if parsed.short_description or parsed.long_description:
            function_description = parsed.short_description or parsed.long_description
        else:
            function_description = None

        args, varargs, kwargs = inspect.getargs(extractor_function.__code__)
        if varargs:
            errors.append("Varargs are not supported")
        if kwargs:
            errors.append("Kwargs are not supported")

        feature_extractor_columns = []
        for param in parsed.params:
            param_errors = []
            arg_map = {
                "name": "",
                "description": "",
                "type": ""
            }
            if param.arg_name is not None:
                arg_map['name'] = param.arg_name
            else:
                param_errors.append("Feature Extractor Parameter name not provided")
            if param.description is not None:
                arg_map['description'] = param.description
            else:
                param_errors.append("Feature Extractor Parameter description not provided")
            if param.type_name is not None:
                arg_map['type'] = param.type_name
            else:
                param_errors.append("Feature Extractor Parameter type not provided")

            feature_extractor_columns.append(
                FeatureExtractorColumn(
                    name = arg_map['name'],
                    data_type = arg_map['type'],
                    description = arg_map['description'],
                    errors = param_errors
                )
            )

        feature_extractor_metadata = FeatureExtractorMetadata(
            feature_extractor_name = extractor.name,
            feature_extractor_description = function_description,
            feature_extractor_columns = feature_extractor_columns,
            feature_extractor_entrypoint = extractor.entrypoint,
            errors = toplevel_errors + errors
        )

        extractors.append(feature_extractor_metadata)

    if len(extractors) == 0:
        return FeatureExtractorMetadata(
            feature_extractor_name = None,
            feature_extractor_description = None,
            feature_extractor_columns = None,
            feature_extractor_entrypoint = None,
            errors = toplevel_errors
        )

    elif len(extractors) > 1:
        toplevel_errors.append("Unable to upload multiple Feature Extractors at once.")

        return FeatureExtractorMetadata(
            feature_extractor_name = None,
            feature_extractor_description = None,
            feature_extractor_columns = None,
            feature_extractor_entrypoint = None,
            errors = toplevel_errors
        )

    elif len(extractors) == 1:
        return extractors.pop()

if __name__ == '__main__':
    import traceback

    try:
        extractor_metadata = generate_metadata()
        if extractor_metadata:
            print_result(generate_metadata().to_json())
    except Exception:
        print_result(json.dumps(dict(errors=[traceback.format_exc()])))
        exit(1)



from docstring_parser import parse
from typing import Dict, Callable, Any

CONVERTIBLE_TYPES = [int, str, float, bool]
EXTRACTOR_TYPES = ["int", "str", "float", "bool"]

def to_convertible(extractor_type):
    if extractor_type == 'int':
        return int
    if extractor_type == 'str':
        return str
    if extractor_type == 'float':
        return float
    if extractor_type == 'bool':
        return bool

def convert(c):
    def do_convert(raw: Dict[str, Any], column):
        value = raw[column]
        if c in [str, float]:
            return c(value)
        if c == int:
            return int(float(value))
        elif c == bool:
            return DataType.BOOL.coerce(value)

    return do_convert

def get_input_type_map(entrypoint_fn):
    docs = entrypoint_fn.__doc__
    parsed = parse(docs)

    input_type_map = {}

    for param in parsed.params:
        column_name = param.arg_name
        type_name = param.type_name
        if type_name in EXTRACTOR_TYPES:
            input_type_map[column_name] = to_convertible(type_name).__name__
        else:
            raise Exception(f"Unsupported input type: {type_name}")

    return input_type_map

def get_input_converter(entrypoint_fn) -> Dict[str, Callable[[str], Any]]:
    docs = entrypoint_fn.__doc__
    parsed = parse(docs)

    converter = {}

    for param in parsed.params:
        column_name = param.arg_name
        type_name = param.type_name

        if type_name in EXTRACTOR_TYPES:
            converter[column_name] = convert(to_convertible(type_name))
        else:
            raise Exception(f"Unsupported input type: {type_name}")

    return converter



from typing import List
from pathlib import Path

import json
import importlib

def import_workdir() -> List[str]:
    errors = []
    for file in Path(".").iterdir():
        if file.suffix == ".py":
            file_name = file.name[: -len(".py")]
            if "." in file_name:
                errors.append(f"Python files cannot contain dots: {file_name}")
                continue
            try:
                importlib.import_module(file_name)
            except Exception as ex:
                errors.append(ex)
    return errors

def print_result(msg):
    print(f"[result] {msg}")

def invalid_extractor(code, error):
    raise InvalidExtractor(code, error)

def deser_output(output: str):
    prefix = "[result]"
    processed = ""
    for line in output.splitlines():
        if line.startswith(prefix):
            processed += line[len(prefix) :]
    try:
        return json.loads(processed)
    except json.decoder.JSONDecodeError:
        raise Exception(f"Could not parse JSON: {output}")

def print_error(message: str, data=None):
    error = {"has_error": True, "error": message, "data": data}
    print_result(json.dumps(error))

class InvalidExtractor(Exception):
    def __init__(self, key, message):
        super().__init__(f"{message} [{key}]")



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['lhub_extractors']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<7.1',
 'dataclasses-json',
 'docstring-parser>=0.1,<0.2',
 'wheel>=0.32.3,<0.33.0']

setup_kwargs = {
    'name': 'lhub-extractors',
    'version': '0.4.0',
    'description': 'LogicHub Feature Extractors',
    'long_description': None,
    'author': 'Bernie Liu',
    'author_email': 'bernie@logichub.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
